/*
 * can_ioctl - can4linux CAN driver module
 *
 * can4linux -- LINUX CAN device driver source
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 * 
 * Copyright (c) 2001 port GmbH Halle/Saale
 * (c) 2001 Heinz-Jrgen Oertel (oe@port.de)
 *          Claus Schroeter (clausi@chemie.fu-berlin.de)
 * (c) 2003 Phil Wilshire (philwil@sysdcs.com)

 *------------------------------------------------------------------
 * $Header: $
 *
 *--------------------------------------------------------------------------
 *
 *
 * modification history
 * --------------------
 * $Log: can_ioctl.c,v $
 *
 * added 5282 buffer handlers
 *  BuffMode sets up buffers after a reset
 *  allow rtr code to be preset
 *--------------------------------------------------------------------------
 */


/**
* \file can_ioctl.c
* \author Heinz-Jrgen Oertel, port GmbH
* $Revision: 1.5 $
* $Date: 2002/10/25 10:39:54 $
*
*
*/

#include "can_defs.h"

int can_Command(struct inode *inode, int cmd);
int can_Send(struct inode *inode, canmsg_t *Tx);
int can_Receive(struct inode *inode, canmsg_t *Rx);
int can_GetStat(struct inode *inode, CanStatusPar_t *s);
// PSW changed this
//int can_Config(struct inode *inode, 
//	       struct file * file,
//	       int target, unsigned long val1,
//	       unsigned long val2);
int can_Config(	struct inode *inode,struct file * file,Command_par_t * argp);

//
// PSW thees are the buffers selected fro use with any instance
// of open
//
extern int Can_SetSelectBufs(int minor, struct file * file,
			     unsigned int * val1, unsigned int * val2);
extern int Can_GetSelectBufs(int minor, struct file * file,
			     unsigned int * val1, unsigned int * val2);

/***************************************************************************/
/**
*
\brief int ioctl(int fd, int request, ...);
the CAN controllers control interface
\param fd The descriptor to change properties
\param request special configuration request
\param ...  traditional a \a char *argp

The \a ioctl function manipulates the underlying device
parameters of the CAN special device.
In particular, many operating characteristics of
character CAN driver may be controlled with \a ioctl requests.
The argument \a fd must be an open file descriptor.

An ioctl request has encoded in it whether the argument is
an \b in parameter or \b out parameter,
and the size of the argument argp in bytes.
Macros and defines used in specifying an \a ioctl request
are located  in  the  file can4linux.h .

The following \a requests are defined:

\li \c COMMAND some commands for
start, stop and reset the CAN controller chip
\li \c CONFIG configure some of the device properties
like acceptance filtering, bit timings, mode of the output control register
or the optional software message filter configuration(not implemented yet).
\li \c STATUS request the CAN controllers status
\li \c SEND a single message over the \a ioctl interface 
\li \c RECEIVE poll a receive message
\li \c CONFIGURERTR configure automatic rtr responses(not implemented)

The third argument is a parameter structure depending on the request.
These are
\code
struct Command_par
struct Config_par
struct CanSja1000Status_par
struct ConfigureRTR_par
struct Receive_par
struct Send_par
\endcode
described in can4linux.h

\par Acceptance Filtering

\b Basic \b CAN.
In the case of using standard identifiers in Basic CAN mode
for receiving CAN messages
only the low bytes are used to set acceptance code and mask
for bits ID.10 ... ID.3

\par
\b PeliCAN.
For acceptance filtering the entries \c AccCode and \c AccMask are used
like specified in the controllers manual for
\b Single \b Filter \b Configuration .
Both are 4 byte entries.
In the case of using standard identifiers for receiving CAN messages
also all 4 bytes can be used.
In this case two bytes are used for acceptance code and mask
for all 11 identifier bits plus additional the first two data bytes.
The SJA1000 is working in the \b Single \b Filter \ Mode .
\code
       Bits
 mask  31 30 .....           4  3  2  1  0
 code
 -------------------------------------------
 ID    28 27 .....           1  0  R  +--+-> unused
                                   T
                                   R
\endcode

\returns
On success, zero is returned.
On error, -1 is returned, and errno is set appropriately.

\par Example
\code
Config_par_t  cfg;
volatile Command_par_t cmd;


    cmd.cmd = CMD_STOP;
    ioctl(can_fd, COMMAND, &cmd);

    cfg.target = CONF_ACCM; 
    cfg.val    = acc_mask;
    ioctl(can_fd, CONFIG, &cfg);
    cfg.target = CONF_ACCC; 
    cfg.val    = acc_code;
    ioctl(can_fd, CONFIG, &cfg);

    cmd.cmd = CMD_START;
    ioctl(can_fd, COMMAND, &cmd);

\endcode

*/
int can_ioctl( __LDDK_IOCTL_PARAM )
{
void *argp;
int retval = -EIO;
//PSW  
//void * argp;
Message_par_t Message;
Command_par_t Command;

  argp = &Message;


  DBGin("can_ioctl");
  DBGprint(DBG_DATA,("cmd=%d", cmd));
  Can_errno = 0;
  
  switch(cmd){

        case COMMAND:
	  if( verify_area(VERIFY_READ, (void *)arg, sizeof(Message_par_t))) {
	     DBGout(); return(retval); 
	  }
	  if( verify_area(VERIFY_WRITE, (void *)arg, sizeof(Message_par_t))) {
	     DBGout(); return(retval); 
	  }
	  //argp = (void *) kmalloc( sizeof(Command_par_t) +1 , GFP_KERNEL );
	  argp = &Command;
	  __lddk_copy_from_user( (void *) argp,(Command_par_t *) arg,
	  					sizeof(Command_par_t));
	  ((Command_par_t *) argp)->retval =
	  		can_Command(inode, ((Command_par_t *) argp)->cmd );
	  ((Command_par_t *) argp)->error = Can_errno;
	  __lddk_copy_to_user( (Command_par_t *)arg, (void *)argp,
	  					sizeof(Command_par_t));
	  //kfree(argp);
	  break;
      case CONFIG:
	  if( verify_area(VERIFY_READ, (void *) arg, sizeof(Command_par_t))) {
	     DBGout(); return(retval); 
	  }
	  if( verify_area(VERIFY_WRITE, (void *) arg, sizeof(Command_par_t))) {
	     DBGout(); return(retval); 
	  }
	  //PSW argp = (void *) kmalloc( sizeof(Config_par_t) +1 ,GFP_KERNEL);
	  argp = &Command;
	  __lddk_copy_from_user( (void *) argp,(Command_par_t *) arg,
	  					sizeof(Command_par_t));
	  ((Command_par_t *) argp)->retval =
	    can_Config(inode, file, (Command_par_t *) argp);

	  ((Command_par_t *) argp)->error = Can_errno;
	  __lddk_copy_to_user( (Command_par_t *) arg, (void *) argp,
	  					sizeof(Command_par_t));
	  //PSW kfree(argp);
	  break;
      case SEND:
	  if( verify_area(VERIFY_READ, (void *) arg, sizeof(Message_par_t))) {
	     DBGout(); return(retval); 
	  }
	  if( verify_area(VERIFY_WRITE, (void *) arg, sizeof(Message_par_t))) {
	     DBGout(); return(retval); 
	  }
	  //PSW argp = (void *)kmalloc( sizeof(Send_par_t) +1 ,GFP_KERNEL );
	  argp = &Message;
	  __lddk_copy_from_user( (void *) argp, (Message_par_t *)arg,
	  				sizeof(Message_par_t));
	  ((Message_par_t *) argp)->retval =
	  		can_Send(inode, &((Message_par_t *) argp)->Tx );
	  ((Message_par_t *) argp)->error = Can_errno;
	  __lddk_copy_to_user( (Message_par_t *) arg, (void *)argp,
	  				sizeof(Message_par_t));
	  //PSW kfree(argp);
	  break;
      case RECEIVE:
	  if( verify_area(VERIFY_READ, (void *) arg, sizeof(Message_par_t))) {
	     DBGout(); return(retval); 
	  }
	  if( verify_area(VERIFY_WRITE, (void *) arg, sizeof(Message_par_t))) {
	     DBGout(); return(retval); 
	  }
	  // PSWargp = (void *)kmalloc( sizeof(Receive_par_t) +1 ,GFP_KERNEL );
	  argp = &Message;
	  __lddk_copy_from_user( (void *)argp, (Message_par_t *)arg,
	  				sizeof(Message_par_t));
	  ((Message_par_t *) argp)->retval =
	  		can_Receive(inode, &((Message_par_t *) argp)->Tx);
	  ((Message_par_t *) argp)->error = Can_errno;
	  __lddk_copy_to_user( (Message_par_t *)arg, (void *) argp,
	  				sizeof(Message_par_t));
	  // PSW kfree(argp);
	  break;

      case STATUS:
  	  printk(" PSW TODO fix the status message \n");
	  if( verify_area(VERIFY_READ, (void *) arg,
	  				sizeof(CanStatusPar_t))) {
	     DBGout(); return(retval); 
	  }
	  if( verify_area(VERIFY_WRITE, (void *) arg,
	  			sizeof(CanStatusPar_t))) {
	     DBGout(); return(retval); 
	  }
	  argp = (void *)kmalloc( sizeof(CanStatusPar_t) +1 ,GFP_KERNEL );

	  ((CanStatusPar_t *) argp)->retval =
	  		can_GetStat(inode, ((CanStatusPar_t *)argp));
	  __lddk_copy_to_user( (CanStatusPar_t *)arg, (void *) argp,
	  				sizeof(CanStatusPar_t));
	  kfree(argp);
	  break;

#ifdef CAN_RTR_CONFIG
      case CONFIGURERTR:
	  if( verify_area(VERIFY_READ, (void *) arg,
	  			sizeof(Message_par_t))){
	     DBGout(); return(retval); 
	  }
	  if( verify_area(VERIFY_WRITE, (void *) arg,
	  			sizeof(Message_par_t))){
	     DBGout(); return(retval); 
	  }
          // PSW let use the stack here ...
          // Make sendPar RxPar and RTRPar all the same
	  //argp =(void *)kmalloc( sizeof(ConfigureRTR_par_t) +1 ,GFP_KERNEL );
	  argp = &Message;
	  __lddk_copy_from_user( (void *) argp,(Message_par_t *) arg,
	  			sizeof(Message_par_t));
	  ((Message_par_t *) argp)->retval =
	  	can_ConfigureRTR(inode,
	  			((Message_par_t *) argp)->message, 
				((Message_par_t *) argp)->Tx );
	  ((Message_par_t *) argp)->error = Can_errno;
	  __lddk_copy_to_user( (Message_par_t *) arg, (void *) argp,
	  			sizeof(Message_par_t));
	  //kfree(argp);
	  break;

      case RTRRESPONSE:
	  if( verify_area(VERIFY_READ, (void *) arg,
	  			sizeof(Message_par_t))){
	     DBGout(); return(retval); 
	  }
	  if( verify_area(VERIFY_WRITE, (void *) arg,
	  			sizeof(Message_par_t))){
	     DBGout(); return(retval); 
	  }
          // PSW let use the stack here ...
          // Make sendPar RxPar and RTRPar all the same
	  //argp =(void *)kmalloc( sizeof(ConfigureRTR_par_t) +1 ,GFP_KERNEL );
	  argp = &Message;
	  __lddk_copy_from_user( (void *) argp,(Message_par_t *) arg,
	  			sizeof(Message_par_t));
	  ((Message_par_t *) argp)->retval =
	  	can_RTRresponse(inode,
	  			((Message_par_t *) argp)->message, 
				((Message_par_t *) argp)->Tx );
	  ((Message_par_t *) argp)->error = Can_errno;
	  __lddk_copy_to_user( (Message_par_t *) arg, (void *) argp,
	  			sizeof(Message_par_t));
	  //kfree(argp);
	  break;

#endif  	/* CAN_RTR_CONFIG */
  
      default:
        DBGout();
	return -EINVAL;
    }
    DBGout();
    return 1;
}



#ifndef DOXYGEN_SHOULD_SKIP_THIS


/* 
* ioctl functions are following here 
*/


int can_Command(struct inode *inode, int cmd)
{
unsigned int minor = MINOR(inode->i_rdev);

    DBGprint(DBG_DATA,("cmd=%d", cmd));
    switch (cmd) {
      case CMD_START:
	    CAN_StartChip(minor);
	    break;
      case CMD_STOP:
	    CAN_StopChip(minor);
	    break;
      case CMD_RESET:
	    CAN_ChipReset(minor);
	    break;
      default:
	    DBGout();
	    return(-EINVAL);
    }
    return 0;
}

/* is not very useful! use it if you are sure the tx queu is empty */
// PSW we will use this to set up RTR response perhaps ??

int can_Send(struct inode *inode, canmsg_t *Tx)
{
unsigned int minor = MINOR(inode->i_rdev);	
canmsg_t tx;

    if( verify_area(VERIFY_READ,Tx,sizeof(canmsg_t)) ) {
	    return -EINVAL;
    }
    __lddk_copy_from_user((canmsg_t *) &tx, (canmsg_t *) Tx,sizeof(canmsg_t));
    return CAN_SendMessage(minor, &tx);
}


int can_Receive(
	struct inode *inode,
	canmsg_t *Rx
	)
{

unsigned int minor = MINOR(inode->i_rdev);
canmsg_t rx;
int len;

    len = CAN_GetMessage(minor,&rx);

    if( len > 0 ){
       /* printk("\nrx[ ] got id 0x%x len %d adr00x%lx\n",rx.id,rx.length,Rx); */
       if( verify_area(VERIFY_WRITE,Rx,sizeof(canmsg_t) ) ) {
	    return -EINVAL;
       }
       __lddk_copy_to_user((canmsg_t *) Rx, (canmsg_t *) &rx,
       						sizeof(canmsg_t));
       return rx.length;
    } else {
     /* no message availiable */
     return 0;
    }
}


int can_Config(
	struct inode *inode,
	struct file * file,
	Command_par_t * argp
	)
{
  
				   
int target         = argp->target;
unsigned long val1 = argp->val1;
unsigned long val2 = argp->val2;
unsigned int minor = MINOR(inode->i_rdev);	

    switch(target) {

      case CONF_ACC:
    DBGprint(DBG_DATA,("mask = 0x%lx, code = 0x%lx", val1, val2));
	   AccMask[minor] = val1;
	   AccCode[minor] = val2;
	   CAN_SetMask(minor, AccCode[minor], AccMask[minor]);		
	   break;

      case CONF_ACC14:
    DBGprint(DBG_DATA,("mask14 = 0x%lx, code14 = 0x%lx", val1, val2));
	   AccMask14[minor] = val1;
	   AccCode14[minor] = val2;
	   CAN_SetMask14(minor, AccCode14[minor], AccMask14[minor]); 
	   break;

      case CONF_ACC15:
    DBGprint(DBG_DATA,("mask15 = 0x%lx, code15 = 0x%lx", val1, val2));
	   AccMask15[minor] = val1;
	   AccCode15[minor] = val2;
	   CAN_SetMask15(minor, AccCode15[minor], AccMask15[minor]);           
	   break;


      case CONF_ACCM:
    DBGprint(DBG_DATA,("acc_mask=0x%lx", val1));
	   AccMask[minor] = val1;
	   CAN_SetMask(minor, AccCode[minor], AccMask[minor]);		
	   break;

      case CONF_ACCM14:
    DBGprint(DBG_DATA,("acc_mask14=0x%lx", val1));
	   AccMask14[minor] = val1;
	   CAN_SetMask14(minor, AccCode14[minor], AccMask14[minor]);      
	   break;

      case CONF_ACCM15:
    DBGprint(DBG_DATA,("acc_mask15=0x%lx", val1));
	   AccMask15[minor] = val1;
	   CAN_SetMask(minor, AccCode15[minor], AccMask15[minor]);	    
	   break;

      case CONF_ACCC:
    DBGprint(DBG_DATA,("acc_code=0x%lx", val1));
	   AccCode[minor] = val1;
	   CAN_SetMask(minor, AccCode[minor], AccMask[minor]);		
	   break;

      case CONF_ACCC14:
    DBGprint(DBG_DATA,("acc_code14=0x%lx", val1));
	   AccCode14[minor] = val1;
	   CAN_SetMask(minor, AccCode14[minor], AccMask14[minor]);	
	   break;

      case CONF_ACCC15:
    DBGprint(DBG_DATA,("acc_code15=0x%lx", val1));
	   AccCode15[minor] = val1;
	   CAN_SetMask15(minor, AccCode15[minor], AccMask15[minor]);   	
	   break;

      case CONF_BUFFM:
    DBGprint(DBG_DATA,("val1 = 0x%lx, val2 = 0x%lx", val1, val2));
           if ( val1 < 16 ) {
	      CAN_SetBuffMode(minor,val1,val2);
	   } else {
	     printk ( " PSW invalid buffer selected %d \n",val1);
           } 
	   break;

      case CONF_BUFFMASK:
    DBGprint(DBG_DATA,("val1 = 0x%lx, val2 = 0x%lx", val1, val2));
           if ( val1 < 16 ) {
	      CAN_SetBuffMask(minor,val1,val2);
	   } else {
	     printk ( " PSW invalid buffer selected %d \n",val1);
           } 
	   break;

      case CONF_BUFFMASKE:
    DBGprint(DBG_DATA,("val1 = 0x%lx, val2 = 0x%lx", val1, val2));
           if ( val1 < 16 ) {
	      CAN_SetBuffMaskE(minor,val1,val2);
	   } else {
	     printk ( " PSW invalid buffer selected %d \n",val1);
           } 
	   break;

      case CONF_TIMING:
	   Baud[minor] = val1;
	   CAN_SetTiming(minor,(int) val1);   
	   break;                    
      case CONF_OMODE:
	   CAN_SetOMode( minor, (int) val1);
	   break;			
#if CAN_USE_FILTER
      case CONF_FILTER:
	   Can_FilterOnOff( minor, (int) val1 );
	   break;
      case CONF_FENABLE:
	   Can_FilterMessage( minor, (int) val1, 1);
	   break;
      case CONF_FDISABLE:
	   Can_FilterMessage( minor, (int) val1, 0);


#endif
      case CONF_SETSELBUFS:
	   Can_SetSelectBufs( minor, file,
			      (unsigned int *)&argp->val1,
			      (unsigned int *)&argp->val2); 
			     
	   break;

      case CONF_GETSELBUFS:
	   Can_GetSelectBufs( minor, file,
			      (unsigned int *)&argp->val1,
			      (unsigned int *)&argp->val2); 
			     
	   break;

      default:
	    DBGout();
	    return(-EINVAL);
    }
    return 0;
}
#endif /* DOXYGEN_SHOULD_SKIP_THIS */
